package cz.cvut.kbss.ear.eshop.service;

import cz.cvut.kbss.ear.eshop.config.RestConfig;
import cz.cvut.kbss.ear.eshop.environment.Generator;
import cz.cvut.kbss.ear.eshop.environment.config.TestSecurityConfig;
import cz.cvut.kbss.ear.eshop.model.Order;
import cz.cvut.kbss.ear.eshop.model.Role;
import cz.cvut.kbss.ear.eshop.model.User;
import cz.cvut.kbss.ear.eshop.security.SecurityUtils;
import cz.cvut.kbss.ear.eshop.security.model.UserDetails;
import org.junit.Ignore;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.test.context.ContextConfiguration;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import static org.junit.Assert.*;

@Ignore
@ContextConfiguration(classes = {TestSecurityConfig.class, RestConfig.class})
public class SecurityTest extends BaseServiceTestRunner {

    @PersistenceContext
    private EntityManager em;

    @Autowired
    private UserService userService;

    @Autowired
    private OrderService orderService;

    @Test(expected = AccessDeniedException.class)
    public void deleteUserThrowsUnauthorizedExceptionForNonAdmin() {
        final User data = Generator.generateUser();
        data.setRole(Role.USER);
        em.persist(data);
        final User user = Generator.generateUser();
        user.setRole(Role.USER);
        SecurityUtils.setCurrentUser(new UserDetails(user));

        try {
            userService.deleteUser(data.getId());
        } finally {
            assertNotNull(em.find(User.class, data.getId()));
        }
    }

    @Test
    public void deleteUserWorksForAdminUser() {
        final User data = Generator.generateUser();
        data.setRole(Role.USER);
        em.persist(data);
        final User user = Generator.generateUser();
        user.setRole(Role.ADMIN);
        SecurityUtils.setCurrentUser(new UserDetails(user));

        userService.deleteUser(data.getId());
        assertNull(em.find(User.class, data.getId()));
    }

    @Test
    public void findAllReturnsOrdersOfUserForRegularUser() {
        final User user = Generator.generateUser();
        user.setRole(Role.USER);
        em.persist(user);
        SecurityUtils.setCurrentUser(new UserDetails(user));

        final List<Order> allOrders = generateOrders(user);
        final List<Order> result = orderService.findAll();
        final List<Order> expected  = allOrders.stream().filter(o -> o.getCustomer().getId().equals(user.getId())).collect(
                Collectors.toList());
        assertEquals(expected.size(), result.size());
        for (Order o : result) {
            assertTrue(expected.stream().anyMatch(exp -> exp.getId().equals(o.getId())));
        }
    }

    private List<Order> generateOrders(User user) {
        final User otherUser = Generator.generateUser();
        user.setRole(Role.USER);
        em.persist(otherUser);
        final List<Order> orders = new ArrayList<>(20);
        for (int i = 0; i < 20; i++) {
            final Order order = new Order();
            order.setCreated(LocalDateTime.now());
            order.setCustomer(Generator.randomBoolean() ? user : otherUser);
            em.persist(order);
            orders.add(order);
        }
        return orders;
    }

    @Test
    public void findAllReturnsAllOrdersForAdminUser() {
        final User user = Generator.generateUser();
        user.setRole(Role.ADMIN);
        em.persist(user);
        SecurityUtils.setCurrentUser(new UserDetails(user));

        final List<Order> allOrders = generateOrders(user);
        final List<Order> result = orderService.findAll();
        assertEquals(allOrders.size(), result.size());
    }
}
